/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 The "llist" package: utility functions for the general handling of
 double-linked lists (in particular, circular lists with a dummy node
 as root).

 The user should supply:
 - the definition of a "Node", typedef'ing "Node" to an appropriate
   structure containing the user data and two links, having name
   "prev" and "next", defined as two pointers to Node's; this
   definition has to appear in an include file called "node.h".
 - A "root" Node, to which the whole linked list will be appended,
   with a definition like "Node  root;".  In addition, the user has to
   issue a call lInit(root); to properly initialize this root Node.

 Every Node to be inserted in the linked list should be appropriately
 allocated with "malloc(sizeof(Node));", and its data part has to be
 correctly filled.
 
 To insert that Node at the beginning or at the end of the linked
 list, the user should call insBegin or insEnd giving as parameters a
 pointer to the root Node and a pointer to the new Node.  To insert
 the new Node before or after another existing Node, the user has to
 call insBefore or insAfter, giving as parameters two pointers to the
 already existing Node ant to the new one.

 The linked list may be sorted calling lSort and giving as parameters
 the pointer to the root Node and an user defined comparison function
 in the style of qsort.

 lRemove may be used to remove a Node from the linked list and to free
 the related memory.  It has to be an actual Node, *NOT* the (dummy)
 root Node.

 lForward and lBackward travel through the whole list, calling a
 specific function for every Node; they take as parameter a pointer to
 the root Node and the name function.

 lClear(&root) returns the root to the state of an empty list, and
 frees all the Nodes of the list.

 $Id: llist.c,v 1.2 2000/02/05 15:30:56 loreti Exp $

 *------------------------------------------------------*/

#include <stdio.h>              /* Standard library */
#include <stdlib.h>
#include "node.h"               /* Local includes */
#include "llist.h"

/**
 | Local procedure prototypes
**/

static Node *lMerge(Node *, Node *, int (*)(const Node *, const Node *));
static Node *lMergeSort(Node *, int (*)(const Node *, const Node *));
static Node *lSplit(Node *, int (*)(const Node *, const Node *));

void insAfter(
  Node *target,                 /* "Old" Node */
  Node *newNode                 /* Node to be inserted after "target" */
) {
  newNode->prev = target;
  newNode->next = target->next;
  target->next->prev = newNode;
  target->next = newNode;
}

void insBefore(
  Node *target,                 /* "Old" Node */
  Node *newNode                 /* Node to be inserted before "target" */
) {
  newNode->next = target;
  newNode->prev = target->prev;
  target->prev->next = newNode;
  target->prev = newNode;
}

void insBegin(
  Node *pRoot,                  /* Root Node */
  Node *newNode                 /* Node to insert at the beginning */
) {
  newNode->next = pRoot->next;
  newNode->prev = pRoot;
  pRoot->next->prev = newNode;
  pRoot->next = newNode;
}

void insEnd(
  Node *pRoot,                  /* Root Node */
  Node *newNode                 /* Node to be inserted at the end */
) {
  newNode->next = pRoot;
  newNode->prev = pRoot->prev;
  pRoot->prev->next = newNode;
  pRoot->prev = newNode;
}

void lBackward(
  const Node *pRoot,
  void (*fun)(const Node *)
) {
  const Node *pNode;

  for (pNode = pRoot->prev;   pNode != pRoot;   pNode = pNode->prev) {
    fun(pNode);
  }
}

void lClear(
  Node *pRoot
) {
  Node *pNode, *p;

  for (pNode = pRoot->next;   pNode != pRoot;   pNode = p) {
    p = pNode->next;
    free(pNode);
  }
  lInit(pRoot);
}

void lForward(
  const Node *pRoot,
  void (*fun)(const Node *)
) {
  const Node *pNode;

  for (pNode = pRoot->next;   pNode != pRoot;   pNode = pNode->next) {
    fun(pNode);
  }
}

void lInit(
  Node *pRoot
) {
  pRoot->next = pRoot->prev = pRoot;
}

void lRemove(
  Node *pNode
) {
  pNode->prev->next = pNode->next;
  pNode->next->prev = pNode->prev;
  free(pNode);
}

void lSort(
  Node *pRoot,
  int (*cmp)(const Node *, const Node *)
) {

/**
 | The linked list is sorted along the "next" links; to simplify the
 | test for the end of the list, the "next" pointer of the last node
 | is temporarily set to NULL.  After the sorting, the "prev" links
 | are set (traveling forward in the list itself) and the "next" link
 | of the last node is restored.
**/

  if (pRoot->next != pRoot->prev) {
    Node *pNode;

    pRoot->prev->next = 0;
    pRoot->next = lMergeSort(pRoot->next, cmp);

    for (pNode = pRoot;   pNode->next != 0;   pNode = pNode->next) {
      pNode->next->prev = pNode;
    }
    pNode->next = pRoot;
    pRoot->prev = pNode;
  }
}

static Node *lMerge(
  Node *list1,
  Node *list2,
  int (*cmp)(const Node *, const Node *)
) {
  Node *pNewList;
  Node **p;

  for (pNewList = 0,  p = &pNewList;   list1 != 0  && list2 != 0;
       p = &((*p)->next)) {
    if (cmp(list1, list2) < 0) {
      *p = list1;
      list1 = list1->next;
    } else {
      *p = list2;
      list2 = list2->next;
    }
  }

  *p = list1 ? list1 : list2;

  return pNewList;
}

static Node *lMergeSort(
  Node *list,
  int (*cmp)(const Node *, const Node *)
) {

/**
 | Straight implementation of mergesort for linked lists; the list is
 | recursively divided in two halves, that are separately sorted and
 | then merged.  See Knuth, Sedgewick or Cormen for implementation
 | details.
**/

  Node *pNode;

  pNode = lSplit(list, cmp);
  list = lMerge(list, pNode, cmp);
  return list;
}

static Node *lSplit(
  Node *list,
  int (*cmp)(const Node *, const Node *)
) {
  Node *pNode, *pNewList = 0;

  while ((pNode = list->next) != 0) {
    if (cmp(list, pNode) > 0) {
      list->next = pNode->next;
      pNode->next = pNewList;
      pNewList = pNode;
    } else {
      list = pNode;
    }
  }

  if (pNewList != 0) {
    pNode = lSplit(pNewList, cmp);
    pNewList = lMerge(pNewList, pNode, cmp);
  }

  return pNewList;
}
