/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 Test program for the "llist" package; generates random
 lists to be sorted and checked for the right ordering.

 $Id: ltest.c,v 1.1 2000/02/03 15:41:48 loreti Exp $

 *------------------------------------------------------*/

#include <stdio.h>              /* Standard library */
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include "node.h"               /* Local includes */
#include "llist.h"

/**
 | Global variables
**/

Node root;                      /* Linked list root */

/**
 | Local procedure prototypes
**/

static void  check(const Node *, const unsigned int, const time_t);
static int   compare(const Node *, const Node *);
static Node *makeNode(Node *, const int);
static void  print(const Node *);
static void  printList(const Node *);

/**
 | Local global variables
**/

static unsigned int outItems;   /* Needed by print() */

int main(
  int   argc,
  char *argv[]
) {
  unsigned int nNodes;          /* Number of nodes in a linked list */
  unsigned int nTries;          /* Number of runs to be performed */
  unsigned int node, trie;      /* Loop indices */
  time_t       now;             /* Local time (used to seed srand) */

/**
 | Test and decode program parameters
**/

  if (argc < 3) {
    printf("Usage: %s nNodes nTries\n", argv[0]);
    return EXIT_FAILURE;

  } else {
    char *pc;

    nNodes = strtoul(argv[1], &pc, 0);
    if (*pc != '\0'   &&   !isspace((unsigned char) *pc)) {
      printf("Invalid character in nNodes=\"%s\"\n", argv[1]);
      return EXIT_FAILURE;
    }

    nTries = strtoul(argv[2], &pc, 0);
    if (*pc != '\0'   &&   !isspace((unsigned char) *pc)) {
      printf("Invalid character in nTries=\"%s\"\n", argv[2]);
      return EXIT_FAILURE;
    }
  }

/**
 | Initialization
**/

  lInit(&root);
  srand(now = time(0));

/**
 | Loop over different runs:
 | - fills the linked list with random integer values;
 | - sorts the linked list;
 | - tests the linked list for ascending ordering;
 | - removes the list nodes.
**/

  for (trie = 0;   trie < nTries;   trie++) {
    for (node = 1;   node <= nNodes;   node++) {
      Node *newNode = makeNode(&root, rand());
      insEnd(&root, newNode);
    }

    lSort(&root, compare);
    check(&root, trie, now);
    lClear(&root);
  }

  return 0;
}

static void check(
  const Node         *pRoot,
  const unsigned int  trie,
  const time_t        seed
) {
  Node *pNode;

  for (pNode = pRoot->next;   pNode->next != pRoot;   pNode = pNode->next) {
    if (pNode->value > pNode->next->value) {
      printf("*** FAILED *** %d > %d\n", pNode->value, pNode->next->value);
      printf("*** Run nr. %u, random seed = %lu\n", trie, seed);
      printList(pRoot);
      break;
    }
  }
}

static int compare(
  const Node *first,
  const Node *second
) {
  return (first->value - second->value);
}

Node *makeNode(
  Node      *pRoot,
  const int  value
) {

/**
 | Creation of a new node (from a single integer value)
**/

  Node *newNode;

  if ((newNode = malloc(sizeof(Node))) == 0) {
    puts("malloc error");
    lClear(pRoot);
    exit(EXIT_FAILURE);
  }

  newNode->value = value;
  return newNode;
}

static void printList(
  const Node *pRoot
) {
  if (pRoot->next == pRoot) {
    puts("The linked list is empty.");

  } else {
    puts("---------------------------------");
    printf("The first node contains %d; the last node contains %d.\n",
           pRoot->next->value, pRoot->prev->value);

    printf("Forward:");
    outItems = 0;
    lForward(pRoot, print);

    printf("\nBackward:");
    outItems = 0;
    lBackward(pRoot, print);

    puts("");
  }
}

static void print(
  const Node *pNode
) {
  if ((outItems++ % 5) == 0) puts("");
  printf("%15d", pNode->value);
}
